import os
import shutil
import stat
import subprocess
import tempfile

from charmhelpers.core import hookenv
from charmhelpers.core.hookenv import (
    log,
    status_set,
)
from charms.reactive import (
    hook,
    when,
    when_all,
    when_file_changed,
    when_not,
    set_state,
    remove_state,
    only_once,
)
from charms.reactive.helpers import data_changed

SCALEBOT_HOME = "/srv/scalebot"
SCALEBOT_REPO = os.path.join(SCALEBOT_HOME, "repo")
PYTHONPATH = os.path.join(SCALEBOT_REPO, "lib")


@when("extension.available")
def configure_jenkins_jobbuilder(extension):
    status_set("maintenance", "Configuring JobBuilder")
    if not os.path.isdir("/etc/jenkins_jobs"):
        os.mkdir('/etc/jenkins_jobs')

    jj_ini = "/etc/jenkins_jobs/jenkins_jobs.ini"
    if os.path.isfile(jj_ini):
        os.remove(jj_ini)

    config = extension.get_connection_info()
    username = config["admin_username"]
    password = config["admin_password"]

    jenkinsini = open(jj_ini, "w")
    jenkinsini.write("""[jenkins]\nuser=%s\n
password=%s
url=http://localhost:8080
""" % (username, password))
    jenkinsini.close()
    remove_state("scalebot.configured.jobbuilder")
    set_state("scalebot.configured.jobbuilder")


@when("extension.changed")
def unconfigure_jobbuilder():
    remove_state("scalebot.configured.jobbuilder")


@only_once
def configure_scalebot_env():
    ''' Set SCALEBOT_REPO and PYTHONPATH'''
    sbprofile = "/etc/profile.d/scalebot.sh"
    sbrepo = 'export SCALEBOT_REPO="%s"' % SCALEBOT_REPO
    ppath = 'export PYTHONPATH="%s"' % PYTHONPATH

    with open(sbprofile, 'w') as f:
        f.write("#Export Scalebot required paths\n")
        f.write("%s\n" % (sbrepo))
        f.write("%s\n" % (ppath))
    set_state("scalebot.configured.env")


def scalebot_git_config_check():
    required_configs = ['scalebot_config_repo', 'scalebot_config_branch']
    missing_configs = []
    config = hookenv.config()
    for c in required_configs:
        if not config.get(c):
            missing_configs.append(c)
    if len(missing_configs) > 0:
        status_set('blocked',
                   'Unable to continue due to missing configs: %s' %
                   ', '.join(missing_configs))
        return False
    return True


@when('apt.installed.git')
def scalebot_git_clone():
    if not scalebot_git_config_check():
        return

    config = hookenv.config()
    repo = config.get('scalebot_config_repo')
    branch = config.get('scalebot_config_branch')
    ssh_key = config.get('scalebot_private_ssh_key')

    if not data_changed('scalebot.git.config', [repo, branch, ssh_key]):
        log("Repo config has not changed, nothing to do.")
        return True

    log("Scalebot repo config has changed, (re-)cloning.")
    remove_state('scalebot.configured.repo')
    status_set('maintenance', 'Cloning Scalebot repository')

    try:
        shutil.rmtree(SCALEBOT_HOME)
    except FileNotFoundError:
        pass

    os.makedirs(SCALEBOT_HOME)
    shutil.chown(SCALEBOT_HOME, user='ubuntu', group='ubuntu')

    known_hosts = os.path.join(hookenv.charm_dir(), 'files', 'known_hosts.lp')

    git_cmd = '/usr/bin/git clone -b %s %s %s' % \
              (branch, repo, os.path.join(SCALEBOT_HOME, 'repo'))
    if ssh_key:
        keyfile = tempfile.NamedTemporaryFile(mode='w', delete=False)
        keypath = keyfile.name
        log("Writing ssh key to %s" % (keypath))
        shutil.chown(keypath, user='ubuntu', group='ubuntu')
        keyfile.write(ssh_key)
        keyfile.close()
        git_ssh = '"/usr/bin/ssh -v -i %s -o UserKnownHostsFile=%s"' % \
                  (keypath, known_hosts)
        git_cmd = 'GIT_SSH_COMMAND=%s %s' % (git_ssh, git_cmd)

    su_git_cmd = ['/bin/su', '-', 'ubuntu', '-c', git_cmd]
    log("Beginning git clone")
    log(su_git_cmd)
    subprocess.run(su_git_cmd, check=True)

    if ssh_key:
        os.unlink(keypath)

    set_state('scalebot.configured.repo')
    remove_state('scalebot.repo.init')


@when_not('scalebot.repo.init')
def scalebot_repo_init():
    init_path = os.path.join(SCALEBOT_REPO, 'scalebot.d', 'init')
    if os.path.exists(init_path):
        status_set('maintenance', 'Executing %s' % (init_path))
        subprocess.check_call(init_path)
    set_state('scalebot.repo.init')

@hook('config-changed')
def scalebot_git_reconfig():
    config = hookenv.config()
    required_configs = ['scalebot_private_ssh_key', 'scalebot_config_repo',
                        'scalebot_config_branch']
    if any(config.changed(key) for key in required_configs):
        scalebot_git_clone()

@only_once
def create_trigger():
    ''' Create the file 'updatejobs' in /var/lib/jenkins '''

    updatejobs = '/var/lib/jenkins/updatejobs'
    string = 'Update This File to Trigger Git Pull'

    with open(updatejobs, 'a') as f:
        f.write(string)
        f.close()
    shutil.chown(updatejobs, user='jenkins', group='jenkins')

@when_file_changed('/var/lib/jenkins/updatejobs')
def git_pull():
    ''' Pull the git repo when the path above is modified'''
    log("Git Pull Requested")
    if not scalebot_git_clone():
        return

    config = hookenv.config()
    repo = config.get('scalebot_config_repo')
    branch = config.get('scalebot_config_branch')
    ssh_key = config.get('scalebot_private_ssh_key')

    known_hosts = os.path.join(hookenv.charm_dir(), 'files', 'known_hosts.lp')

    git_cmd = '/usr/bin/git -C %s pull' % \
              (os.path.join(SCALEBOT_HOME, 'repo'))
    if ssh_key:
        keyfile = tempfile.NamedTemporaryFile(mode='w', delete=False)
        keypath = keyfile.name
        log("Writing ssh key to %s" % (keypath))
        shutil.chown(keypath, user='ubuntu', group='ubuntu')
        keyfile.write(ssh_key)
        keyfile.close()
        git_ssh = '"/usr/bin/ssh -v -i %s -o UserKnownHostsFile=%s"' % \
                  (keypath, known_hosts)
        git_cmd = 'GIT_SSH_COMMAND=%s %s' % (git_ssh, git_cmd)

    su_git_cmd = ['/bin/su', '-', 'ubuntu', '-c', git_cmd]
    log("pulling git repo")
    log(su_git_cmd)
    subprocess.run(su_git_cmd, check=True)
    if ssh_key:
        os.unlink(keypath)


@when_all('scalebot.configured.repo',
          'scalebot.configured.jobbuilder',
          'scalebot.configured.juju')
def scalebot_ready():
    status_set('active', 'ScaleBot is Ready')
    
@when_all('scalebot.configured.repo',
          'scalebot.configured.jobbuilder',
          'scalebot.configured.env')
def importjenkins():
    config = hookenv.config()
    jobset = config.get('scalebot_jobset')
    jj = os.path.join(hookenv.charm_dir(), 'files', 'jobsync.py')
    subprocess.run([jj, '-j', jobset, '-s'], check=True)


@hook('config-changed')
def scalebot_juju_reconfig():
    config = hookenv.config()
    required_configs = ['scalebot_juju_clouds', 'scalebot_juju_credentials',
                        'scalebot_juju_model_defaults']
    if any(config.changed(key) for key in required_configs):
        scalebot_configure_juju()


@when_all('apt.installed.juju')
def scalebot_configure_juju():
    config = hookenv.config()

    clouds = config.get('scalebot_juju_clouds')
    creds = config.get('scalebot_juju_credentials')
    modeldefaults = config.get('scalebot_juju_model_defaults')
    bootstrapconstraints = config.get('scalebot_juju_bootstrap_constraints')

    if not data_changed('scalebot.juju.config', [clouds, creds]):
        log("Juju config has not changed, nothing to do.")
        return True
    log("Juju config has changed, (re-)bootstrapping.")
    remove_state('scalebot.configured.juju')
    status_set('maintenance', 'Bootstrapping Juju controller')

    # In case we're reconfiguring, destroy any pre-existing
    # controller before re-bootstrapping
    subprocess.run(['/bin/su', '-', 'jenkins', '-c',
                    "juju destroy-controller jenkins -y"],
                   check=False)

    if clouds:
        cloudsfile = tempfile.NamedTemporaryFile(mode='w', delete=False)
        cloudsfile.write(clouds)
        # Make it readable by user jenkins
        os.chmod(cloudsfile.name,
                 stat.S_IRUSR | stat.S_IWUSR | stat.S_IRGRP | stat.S_IROTH)
        cloudsfile.close()
        command = "juju add-cloud --replace scalebot %s" % (cloudsfile.name)
        subprocess.run(['/bin/su', '-', 'jenkins', '-c', command],
                       check=True)
        os.unlink(cloudsfile.name)

    if creds:
        credsfile = tempfile.NamedTemporaryFile(mode='w', delete=False)
        credsfile.write(creds)
        # Make it readable by user jenkins
        os.chmod(credsfile.name,
                 stat.S_IRUSR | stat.S_IWUSR | stat.S_IRGRP | stat.S_IROTH)
        credsfile.close()
        command = "juju add-credential -f %s --replace scalebot" % \
                  (credsfile.name)
        subprocess.run(['/bin/su', '-', 'jenkins', '-c', command],
                       check=True)
        os.unlink(credsfile.name)

    modeldefaultsparam = ""
    if modeldefaults:
        modeldefaultsfile = tempfile.NamedTemporaryFile(mode='w', delete=False)
        modeldefaultsfile.write(modeldefaults)
        os.chmod(modeldefaultsfile.name,
                 stat.S_IRUSR | stat.S_IWUSR | stat.S_IRGRP | stat.S_IROTH)
        modeldefaultsfile.close()
        modeldefaultsparam=" --model-default %s" % (modeldefaultsfile.name)

    bconstraints = ""
    if bootstrapconstraints:
        bconstraints = ' --bootstrap-constraints "%s"' % (bootstrapconstraints)

    subprocess.run(['/bin/su', '-', 'jenkins', '-c',
                    "juju bootstrap --debug scalebot jenkins" + \
                    modeldefaultsparam + bconstraints],
                   check=True)

    if modeldefaults:
        os.unlink(modeldefaultsfile.name)

    set_state('scalebot.configured.juju')

@hook('stop')
def scalebot_juju_destroy_controller():
    subprocess.run(['/bin/su', '-', 'jenkins', '-c',
                    'juju destroy-controller --destroy-all-models -y jenkins'])
